import os
from dataset import *


class Parser:
    '''
    Parser for the dataset (list of interpretations, each one loaded in Sample class).

    Use the method parse() to get dataset.

    Note that the current version supports only constants (functions of arity 0), functions with higher arity are not supported.
    '''

    def __init__(self, pathToDataFile):
        '''
        Constructs new parser. Use parse() method to get dataset.

        :type pathToDataFile: string
        :rtype: Parser
        '''
        self.path = pathToDataFile

    def parse(self):
        '''
        Parses the given file to Dataset.

        Possible exception if the file was not found.

        :rtype: Dataset
        '''
        if not os.path.isfile(self.path):
            raise Exception('The file was not found \'' + self.path + '\'.')

        builder = DatasetBuilder()
        with open(self.path, 'r') as file:
            for line in file:
                if len(line.strip()) < 1:
                    continue
                posClass, cnf = self.parseLine(line, builder.predicateFactory, builder.constantFactory)
                builder.append(Sample(cnf, posClass))

        self.predicateFactory = builder.predicateFactory
        return builder.build()

    def parseLine(self, line, predicateFactory, constantFactory):
        '''
        Parses sample in format 'class : parent(johan,anna),female(marie),male(johan).'.
        Class is either positive '+' or negative '-'. Delimiter ':' splits class and the interpretation which is ended by a dot.

        Note that only function-free ground facts are supported, e.g. all arguments are parsed as constant with zero arity.
        
        :type line: string
        :type predicateFactory: PredicateFactory
        :type constantFactory: ConstantFactory
        :rtype: tuple of class (T|F) and parsed clause
        '''
        classClauseDelimiter = ':'
        endingCharacter = '.'
        negationLiteral = '-'
        openingBracket = '('
        closingBracket = ')'
        separator = ','

        def retrieveArgument(rest):
            if len(rest) < 1:
                raise Exception("Line ended too shortly ('" + rest + "').")
            if closingBracket == rest[0]:
                return None, rest
            if separator == rest[0]:
                rest = rest[1:]
            idx = 0
            brackets = 0
            while True:
                if idx >= len(rest) or brackets < -1:
                    raise Exception("Cannot parse part of line starting with '" + rest + "'.")
                if (separator == rest[idx] or closingBracket == rest[idx]) and brackets <= 0:
                    break
                if openingBracket == rest[idx]:
                    brackets = brackets + 1
                if closingBracket == rest[idx]:
                    brackets = brackets - 1
                idx = idx + 1
            return constantFactory.create(rest[:idx]), rest[idx:].strip()

        classPart = line[:line.find(classClauseDelimiter)].strip()

        positiveClass = True
        if "-" == classPart:
            positiveClass = False
        if classPart not in ["+", "-"]:
            raise Exception("Unknown class token used '" + classPart + "'.")

        rest = line[line.find(classClauseDelimiter) + 1:].strip()

        builder = InterpretationBuilder()
        while True:
            if endingCharacter == rest[0]:
                break
            positive = True
            if negationLiteral == rest[0]:
                positive = False
                rest = rest[1:]
            comma = rest.find(separator)
            opening = rest.find(openingBracket)
            endingDot = rest.find(endingCharacter)

            assert True == positive

            if comma < opening or (-1 == opening and -1 != comma) :
                predicateName = rest[:comma]
                atom = Atom(predicateFactory.create(predicateName, 0), [])
                builder.append(Literal(atom, positive))
                rest = rest[comma + 1:].strip()
                builder.append(Literal(atom, positive))
            elif (endingDot < opening) or (-1 == opening):
                predicateName = rest[:endingDot]
                atom = Atom(predicateFactory.create(predicateName, 0), [])
                builder.append(Literal(atom, positive))
                rest = rest[endingDot + 1:].strip()
                builder.append(Literal(atom, positive))
                break
            else:
                predicateName = rest[:opening]
                rest = rest[opening + 1:]
                arguments = []
                while True:
                    argument, restLine = retrieveArgument(rest)
                    arguments.append(argument)
                    rest = restLine
                    if closingBracket == restLine[0]:
                        rest = restLine[1:]
                        break
                if separator == rest[0]:
                    rest = rest[1:]
                atom = Atom(predicateFactory.create(predicateName, len(arguments)), arguments)
                builder.append(Literal(atom, positive))
        return positiveClass, builder.build()
